/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

module main;

import std.math;
import std.algorithm;
import std.meta;

import dplug.core,
       dplug.client;

import gui;

// PunkLabs
import punklabs;

import dsp_drum;
import dsp_output;

mixin(pluginEntryPoints!OneTrickSimianClient);

enum Module
{
    Output,
    
    /* //Multi Channel Out
    OutputKick,
    OutputSnare,
	OutputRimshot,
	OutputClap,
	OutputTom1,
	OutputTom2,
	OutputTom3,
	OutputHihatClosed,
	OutputHihatOpen,
	OutputCymbal,
	*/
	
    Kick, FirstKitPiece=Kick,
    Snare,
    Rimshot,
    Clap,
    Tom1,
    Tom2,
    Tom3,
    HihatClosed,
    HihatOpen,
    Cymbal,
    
    Count,
    KitPieceCount=Count-1,
}

enum : int	//################## MUST BE IN ORDER OF DEFINITIONS BELOW!!! ##################
{
	// Waveform CC Bug Workaround:
	paramPlaceholder,
	
	// Globals:
	paramGlobalGain,
	paramGlobalDrive,
	paramGlobalSensitivity,
	paramGlobalSaturation,
	paramGlobalReverbDecay,
	paramGlobalReverbGate,

	// Module:
	paramModuleSTART, // = = = = = = = = = = =
   	paramModuleGain = paramModuleSTART,
   	paramModulePan,
   	paramModuleLowpassFreq,
   	paramModuleTone,
   	paramModuleBend,
   	paramModuleDecay,
   	paramModulePunch,
   	paramModuleToneNoise,
   	paramModuleClick,
   	paramModuleGainDynamics,
   	paramModuleBendDynamics,
   	paramModuleLowpassQ,
   	paramModuleSaturation,
   	paramModuleLowpassEnvBend,
   	paramModuleLowpassEnvBendDynamics,
   	paramModuleOscWaveform,
   	paramModuleSendReverb,
   	paramModuleIcon,
   	paramModuleColor,
	paramModuleEND, // = = = = = = = = = = =

	// Metadata
	paramModuleSTRIDE = paramModuleEND-paramModuleSTART,
	paramModuleCOUNT = Module.KitPieceCount,
   	paramMAX = paramModuleSTART + (paramModuleSTRIDE * paramModuleCOUNT),
}

struct ParamDescription
{
public:
nothrow:
@nogc:
	string name = "";
	string suffix = "";
	int type = 0; //Float, Bool, Etc
	FaustParam faustParam;
	this(string name, string suffix, int type, FaustParam faustParam)
	{
		// Duplication isn't needed since we only use constant values here...
		this.name = name;//StringHelper.duplicateString(name);
		this.suffix = suffix;//StringHelper.duplicateString(suffix);
		this.type = type;
		this.faustParam = faustParam;
	}
}

enum ParamType
{
	Dummy,
	Bool,
	Enum,
	Float,
	Gain,
	Integer,
	LinearFloat,
	LogFloat,
	PowFloat,
}

enum Note
{
	Kick = 36,
	Snare = 38,
	Rimshot = 37,
	Clap = 39,
	Tom1 = 41, Tom1Alt = 40,
	Tom2 = 45, Tom2Alt = 43,
	Tom3 = 48, Tom3Alt = 47,
	HihatClosed = 42,
	HihatPedal = 44,
	HihatOpen = 46,
	Cymbal = 49,
	CymbalChoke = 51,
}




final class OneTrickSimianClient : CustomClient
{
public:
nothrow:
@nogc:

    this()
    {
    	allocateBuffers();
    	allocateDSP();
    }
    ~this()
    {
        freeDSP();
        freeBuffers();
    }
    
    void allocateDSP()
	{
		if(!dspModules[0])
		{
			int m = 0;
			dspModules[m++] = mallocNew!(PunkDSP)(mallocNew!(DSP_Output)());

			dspModules[m++] = mallocNew!(PunkDSP)(mallocNew!(DSP_Drum)()); //Kick
			dspModules[m++] = mallocNew!(PunkDSP)(mallocNew!(DSP_Drum)()); //Snare
			dspModules[m++] = mallocNew!(PunkDSP)(mallocNew!(DSP_Drum)()); //Rimshot
			dspModules[m++] = mallocNew!(PunkDSP)(mallocNew!(DSP_Drum)()); //Clap
			dspModules[m++] = mallocNew!(PunkDSP)(mallocNew!(DSP_Drum)()); //Tom1
			dspModules[m++] = mallocNew!(PunkDSP)(mallocNew!(DSP_Drum)()); //Tom2
			dspModules[m++] = mallocNew!(PunkDSP)(mallocNew!(DSP_Drum)()); //Tom3
			
			dspModules[m++] = mallocNew!(PunkDSP)(mallocNew!(DSP_Drum)()); //Hihat Closed
			dspModules[m++] = mallocNew!(PunkDSP)(mallocNew!(DSP_Drum)()); //Hihat Open
			dspModules[m++] = mallocNew!(PunkDSP)(mallocNew!(DSP_Drum)()); //Cymbal
		}

	}
	void freeDSP()
	{
		if(dspModules[0])
		{
			foreach(int i; 0..cast(int)dspModules.length)
			{
				dspModules[i].destroyFree();
				dspModules[i] = null;
			}
		}

	}

    void allocateBuffers()
    {
    	if (stereoHelper[1] is null)
    	{
			//stereoHelper[1] = stereoHelperBuffer.ptr;
			stereoHelper[1] = mallocSliceNoInit!float(maxFramesInProcess()).ptr;
		}
		foreach (int i; 0..3)
		{
			if (dspBufferTemp[i] is null)
				dspBufferTemp[i] = mallocSliceNoInit!float(maxFramesInProcess()).ptr;
			if (dspBufferAccum[i] is null)
				dspBufferAccum[i] = mallocSliceNoInit!float(maxFramesInProcess()).ptr;
		}
    }
    void freeBuffers()
    {
  		if (stereoHelper[1] !is null)
  		{
			stereoHelper[1].destroyFree();
			stereoHelper[1] = null;
		}
		foreach (int i; 0..3)
		{
			if (dspBufferTemp[i])
			{
				dspBufferTemp[i].destroyFree();
				dspBufferTemp[i] = null;
			}
			if (dspBufferAccum[i])
			{
				dspBufferAccum[i].destroyFree();
				dspBufferAccum[i] = null;
			}
		}
	}
    override PluginInfo buildPluginInfo()
    {
        static immutable PluginInfo pluginInfo = parsePluginInfo(import("plugin.json"));
        return pluginInfo;
    }

    override float tailSizeInSeconds()
    {
        return 3.0f; // 2s + 1s reverb...?
    }

	Parameter moduleParam(int moduleParamIndex, int moduleIndex)
	{
		return param(paramModuleSTART + (moduleIndex * paramModuleSTRIDE) + (moduleParamIndex - paramModuleSTART));
	}
    override Parameter[] buildParameters()
    {
    	allocateDSP();
    	
    	foreach(dsp; dspModules)
    		dsp.initializeParams();
    		
		// Bulid Parameter Descriptions. ################## MUST BE IN ORDER OF ENUM ABOVE!!! ##################
		{
			int i=0;
			int dspIndex = 0;
			
			// Globals:
			{
				auto dsp = dspModules[dspIndex++];
				// Waveform CC Bug Workaround:
				paramDesc[i++] = ParamDescription("reserved", "", ParamType.Dummy, dsp.findParam(""));
				paramDesc[i++] = ParamDescription("global_gain", "dB", ParamType.Gain, dsp.findParam("Gain_Global"));
				paramDesc[i++] = ParamDescription("global_drive", "dB", ParamType.Gain, dsp.findParam("Drive_Global"));
				paramDesc[i++] = ParamDescription("global_sensitivity", "%", ParamType.Float, FaustParam("", null, 1, 0, 1, 0.001f)); //dsp.findParam("Sensitivity_Global")
				paramDesc[i++] = ParamDescription("global_saturation", "%", ParamType.Float, dsp.findParam("Saturation_Global"));
				paramDesc[i++] = ParamDescription("global_reverbdecay", "%", ParamType.Float, dsp.findParam("ReverbDecay_Global"));
				paramDesc[i++] = ParamDescription("global_reverbgate", "dB", ParamType.Gain, dsp.findParam("ReverbGate_Global"));
		    }
			
			// Modules:
			int index=0;
			foreach(prefix; AliasSeq!("kick", "snare","rim","clap","tom1","tom2","tom3","hhclosed","hhopen","cymbal"))
			{
				bool isCymbal = (dspIndex > 7);
				auto dsp = dspModules[dspIndex++];
				paramDesc[i++] = ParamDescription(prefix~"_gain", "dB", ParamType.Gain, dsp.findParam("Gain"));
				paramDesc[i++] = ParamDescription(prefix~"_pan", "%", ParamType.Float, dsp.findParam("Pan"));
				paramDesc[i++] = ParamDescription(prefix~"_lowpassfreq", "Hz", ParamType.LogFloat, dsp.findParam("LowpassFreq"));
				paramDesc[i++] = ParamDescription(prefix~"_oscpitch", "Hz", ParamType.LogFloat, dsp.findParam("OscPitch"));
				paramDesc[i++] = ParamDescription(prefix~"_oscbend", "st", ParamType.Float, dsp.findParam("OscEnvBend"));
				paramDesc[i++] = ParamDescription(prefix~"_envdecay", "ms", ParamType.Float, dsp.findParam("EnvDecay"));
				paramDesc[i++] = ParamDescription(prefix~"_envpunch", "%", ParamType.Float, dsp.findParam("EnvPunch"));
				paramDesc[i++] = ParamDescription(prefix~"_tonenoise", "%", ParamType.Float, dsp.findParam("FadeToneNoise"));
				paramDesc[i++] = ParamDescription(prefix~"_click", "%", ParamType.Float, dsp.findParam("FadeClick"));
				paramDesc[i++] = ParamDescription(prefix~"_gaindynamics", "%", ParamType.Float, dsp.findParam("GainDynamics"));
				paramDesc[i++] = ParamDescription(prefix~"_oscbenddynamics", "%", ParamType.Float, dsp.findParam("OscEnvBendDynamics"));
				paramDesc[i++] = ParamDescription(prefix~"_lowpassq", "%", ParamType.Float, dsp.findParam("LowpassQ"));
				paramDesc[i++] = ParamDescription(prefix~"_saturation", "%", ParamType.Float, dsp.findParam("Saturation"));
				paramDesc[i++] = ParamDescription(prefix~"_lowpassbend", "st", ParamType.Float, dsp.findParam("LowpassEnvBend"));
				paramDesc[i++] = ParamDescription(prefix~"_lowpassbenddynamics", "%", ParamType.Float, dsp.findParam("LowpassEnvBendDynamics"));
				auto waveformParam = dsp.findParam("OscWaveform");
				if(isCymbal)
					waveformParam.initial = 100;
				paramDesc[i++] = ParamDescription(prefix~"_oscwaveform", "%", ParamType.Float, waveformParam);
				paramDesc[i++] = ParamDescription(prefix~"_reverb", "%", ParamType.Float, dsp.findParam("SendReverb"));
				const immutable(int)[10] defaultIcons = [0,1,2,3,4,5,6,7,8,7];
				paramDesc[i++] = ParamDescription(prefix~"_icon", "", ParamType.Integer, FaustParam("", null, defaultIcons[index], 0, 8, 1)); //Default to GenericDrum,/Cymbal
				paramDesc[i++] = ParamDescription(prefix~"_color", "", ParamType.Integer, FaustParam("", null, index%8, 0, 7, 1)); //Default to Rainbow
				index++;
			}
		}

        auto params = makeVec!Parameter();
        
        // Add in enum order!
		foreach (int i; 0..paramMAX)
		{
			auto p = paramDesc[i];
			/*
				https://dplug.dpldocs.info/dplug.client.params.html
				BoolParameter
				EnumParameter [undocumented]
				FloatParameter
				GainParameter [-inf to value dB log mapping]
				IntegerParameter
				LinearFloatParameter
				LogFloatParameter [frequency mapping]
				Parameter [base?]
				PowFloatParameter [x^N mapping]
			*/
			switch(p.type)
			{
			case ParamType.Dummy:
				params ~= mallocNew!LinearFloatParameter(i, p.name, p.suffix, 0, 1, 0).nonAutomatable();
				break;
			case ParamType.Integer:
				params ~= mallocNew!IntegerParameter(i, p.name, p.suffix, cast(int)p.faustParam.min, cast(int)p.faustParam.max, cast(int)p.faustParam.initial);
				break;
			case ParamType.Float:
				params ~= mallocNew!LinearFloatParameter(i, p.name, p.suffix, p.faustParam.min, p.faustParam.max, p.faustParam.initial);
				break;
			case ParamType.LogFloat:
				params ~= mallocNew!LogFloatParameter(i, p.name, p.suffix, p.faustParam.min, p.faustParam.max, p.faustParam.initial);
				break;
			case ParamType.Gain:
				params ~= mallocNew!GainParameter(i, p.name, p.faustParam.max, p.faustParam.initial);
				break;
	     	case ParamType.Bool:
				params ~= mallocNew!BoolParameter(i, p.name, p.faustParam.initial != 0);
				break;
			default:
			}

		}
        
        return params.releaseData();
    }

    override LegalIO[] buildLegalIO()
    {
        auto io = makeVec!LegalIO();
        
        // Let's support Mono for convenience...
        io ~= LegalIO(0, 1);
        io ~= LegalIO(0, 2);
        
        //io ~= LegalIO(0, 10); //Multi Channel Out
        
        // Strict IO based on the DSP:
    	//allocateSimian();
        //io ~= LegalIO(dspSimian.getNumInputs(), dspSimian.getNumOutputs());
        
        return io.releaseData();
    }

    // This override is optional, the default implementation will
    // have one default preset.
    override Preset[] buildPresets()
    {
    	auto internalPresets = makeVec!Preset();
    	
    	// Internal Presets
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Basic.preset")));
		    preset.setName("Basic");
		    PresetHelper.applyPreset(this, preset, false);
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Factorial Fiction.preset")));
		    preset.setName("Factorial Fiction");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Passive Aggressive.preset")));
		    preset.setName("Passive Aggressive");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Sequence.preset")));
		    preset.setName("Sequence");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/No Problem.preset")));
		    preset.setName("No Problem");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Bar Fight.preset")));
		    preset.setName("Bar Fight");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Fuchsia Sunrise.preset")));
		    preset.setName("Fuchsia Sunrise");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Fuchsia Sunset.preset")));
		    preset.setName("Fuchsia Sunset");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Sandy Fuego.preset")));
		    preset.setName("Sandy Fuego");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Undeadhead.preset")));
		    preset.setName("Undeadhead");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Italo Disco.preset")));
		    preset.setName("Italo Disco");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Disco Sol.preset")));
		    preset.setName("Disco Sol");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Discopsychology.preset")));
		    preset.setName("Discopsychology");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Clique Beetle.preset")));
		    preset.setName("Clique Beetle");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Suburbanight.preset")));
		    preset.setName("Suburbanight");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Eticroquette.preset")));
		    preset.setName("Eticroquette");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Whatever.preset")));
		    preset.setName("Whatever");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Dustbin.preset")));
		    preset.setName("Dustbin");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Thrashcan.preset")));
		    preset.setName("Thrashcan");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Shhhhh.preset")));
		    preset.setName("Shhhhh");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Aquatic Traffic.preset")));
		    preset.setName("Aquatic Traffic");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Boomsticks.preset")));
		    preset.setName("Boomsticks");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Alpacalypse.preset")));
		    preset.setName("Alpacalypse");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Sasquelch.preset")));
		    preset.setName("Sasquelch");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Life Support.preset")));
		    preset.setName("Life Support");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Hootenanny.preset")));
		    preset.setName("Hootenanny");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Raptor Hatchling.preset")));
		    preset.setName("Raptor Hatchling");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Oh Shoot.preset")));
		    preset.setName("Oh Shoot");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Bubble Factory.preset")));
		    preset.setName("Bubble Factory");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Macrochip.preset")));
		    preset.setName("Macrochip");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Power Up.preset")));
		    preset.setName("Power Up");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Power Down.preset")));
		    preset.setName("Power Down");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Duh Duh Duh.preset")));
		    preset.setName("Duh Duh Duh");
		    internalPresets ~= preset;
		}
    	{
		    Preset preset = PresetHelper.loadBinaryPresetFromMemory(this, cast(ubyte[])(import("Presets/Tommy Tom Tom.preset")));
		    preset.setName("Tommy Tom Tom");
		    internalPresets ~= preset;
		}
		{ // Default
			Preset preset = makeDefaultPreset();
			preset.setName("Init");
	        internalPresets ~= preset;
	    }

		_internalPresetBank = mallocNew!PresetBank(this, internalPresets.releaseData());

    
    	// Public Presets
        auto presets = makeVec!Preset();

		// REQUIRED: Dplug:VST3 can overwrite bank[_current] without safety checks.
		{ // Default
			Preset preset = makeDefaultPreset();
			preset.setName("Init");
	        presets ~= preset;
	    }

        return presets.releaseData();
    }
	override PresetBank internalPresetBank()
	{
		return _internalPresetBank;
	}


    override int maxFramesInProcess()
    {
	    //Dplug automatically splits buffers on MidiEvent bounds, so the main concern is automation resolution.
	    //256 is approx 172hz
        return 256;
    }

    override void reset(double sampleRate, int maxFrames, int numInputs, int numOutputs) 
    {
    	foreach(dsp; dspModules)
    		dsp.initialize(cast(int)sampleRate).trackSilence(100);
    		
        allocateBuffers();
    }

	/* //Stress Test
	void* testPtr = null;
	int testPreset = 0;
	int testPresetDelay = 1000;
	//*/
    override void processAudio(const(float*)[] inputs, float*[]outputs, int frames, TimeInfo info)
    {
		/* // Stress Test
		import core.stdc.stdlib;
		foreach(z; 0..100000)
		{
			if (testPtr !is null)
				free(testPtr);
			testPtr = malloc(z%24+8);
		}
		//*/
		
		/* //Stress Test
		testPresetDelay--;
		if (testPresetDelay <= 0) {
			testPresetDelay = 10;
			testPreset = (testPreset+1) % 30;
			Preset preset = PresetHelper.getFactoryPreset(this, testPreset);
			if (preset !is null)
				PresetHelper.applyPresetFromGUI(this, preset);
		}
		//*/
		

        foreach(MidiMessage msg; getNextMidiMessages(frames)) //TODO: Use ref?
        {
        	/* // We could affect global decay with the Modwheel...?
        	if (msg.isControlChange() && (msg.controlChangeControl() == MidiControlChange.modWheel)) {
        		modWheel = msg.controlChangeValue0to1();
        	} else
        	*/
        	if (msg.isNoteOn()) {
            	auto velocity = cast(float)(msg.noteVelocity()+1.0f)/128.0f;
            	velocity = 1.0f + (velocity-1.0f) * readParam!float(paramGlobalSensitivity);
            	switch(msg.noteNumber())
            	{
            		case Note.Kick:
            			dspModules[Module.Output].wakeUp();
            			dspModules[Module.Kick].wakeUp();
            			dspModules[Module.Kick].trigger(velocity);
	            		break;
            		case Note.Snare:
            			dspModules[Module.Output].wakeUp();
            			dspModules[Module.Snare].wakeUp();
            			dspModules[Module.Snare].trigger(velocity);
	            		break;
            		case Note.Tom3: case Note.Tom3Alt:
            			dspModules[Module.Output].wakeUp();
            			dspModules[Module.Tom3].wakeUp();
						dspModules[Module.Tom3].trigger(velocity);
	            		break;
            		case Note.HihatClosed:
            			dspModules[Module.Output].wakeUp();
            			dspModules[Module.HihatClosed].wakeUp();
            			dspModules[Module.HihatClosed].trigger(velocity);
            			dspModules[Module.HihatOpen].choke();
	            		break;
            		case Note.HihatPedal:
            			dspModules[Module.HihatClosed].choke();
            			dspModules[Module.HihatOpen].choke();
	            		break;
            		case Note.HihatOpen:
            			dspModules[Module.Output].wakeUp();
            			dspModules[Module.HihatOpen].wakeUp();
            			dspModules[Module.HihatOpen].trigger(velocity);
            			dspModules[Module.HihatClosed].choke();
	            		break;
	       			version(Lite) { }
	       			else {
            		case Note.Rimshot:
            			dspModules[Module.Output].wakeUp();
            			dspModules[Module.Rimshot].wakeUp();
            			dspModules[Module.Rimshot].trigger(velocity);
	            		break;
            		case Note.Clap:
            			dspModules[Module.Output].wakeUp();
            			dspModules[Module.Clap].wakeUp();
            			dspModules[Module.Clap].trigger(velocity);
	            		break;
            		case Note.Tom1: case Note.Tom1Alt:
            			dspModules[Module.Output].wakeUp();
            			dspModules[Module.Tom1].wakeUp();
            			dspModules[Module.Tom1].trigger(velocity);
	            		break;
            		case Note.Tom2: case Note.Tom2Alt:
            			dspModules[Module.Output].wakeUp();
            			dspModules[Module.Tom2].wakeUp();
            			dspModules[Module.Tom2].trigger(velocity);
	            		break;
	            	case Note.Cymbal:
            			dspModules[Module.Output].wakeUp();
            			dspModules[Module.Cymbal].wakeUp();
            			dspModules[Module.Cymbal].trigger(velocity);
	            		break;
            		case Note.CymbalChoke:
            			dspModules[Module.Cymbal].choke();
	            		break;
	            	}
	            	default:
            	}
            }
            /* // Not needed...
            else if (msg.isAllNotesOff() || msg.isAllSoundsOff()) {
            	dspKick.choke();
            	dspSnare.choke();
            	dspTom1.choke();
            	dspTom2.choke();
            	dspTom3.choke();
            	dspHihatClosed.choke();
            	dspHihatOpen.choke();
            	dspCymbal.choke();
            }
            */
        }

		// UI Faust Params
		
		foreach (int i; 0..paramMAX)
		{
			auto param = paramDesc[i];
			auto faustParam = param.faustParam;
			if(!faustParam.val)
				continue;
			switch(param.type)
			{
/*
				case ParamType.Integer:
					*(paramDesc[i].faustParam.val) = cast(float)readParam!int(i);
					break;
*/
				case ParamType.Gain:
					*(faustParam.val) = max(readParam!float(i), -100); //-INF can screw up Faust!
					break;
					
				default:
					*(faustParam.val) = readParam!float(i);
			}
		}
		
		if (dspModules[Module.Output].isSilent())
		{
			foreach (int chan; 0..cast(int)outputs.length)
				outputs[chan][0..frames] = 0;
		}
		else
		{
			allocateBuffers();
			
			//Clear
			//dspBufferAccum[0..outputs.length][0..frames] = 0;
			foreach (int chan; 0..3)
				dspBufferAccum[chan][0..frames] = 0;
			
			
			foreach(int m; 0..Module.KitPieceCount)
			{
				auto dsp = dspModules[Module.FirstKitPiece + m];
				if(!dsp.isSilent()) {
					dsp.compute(frames, null, cast(float*[])dspBufferTemp);
					addBuffers(frames, dspBufferAccum, dspBufferTemp);
				}
			}

			switch(outputs.length)
			{
				case 0:
					break;
				case 1:
					allocateBuffers();
					stereoHelper[0] = outputs[0];
				    dspModules[Module.Output].compute(frames, cast(float*[])dspBufferAccum, cast(float*[])stereoHelper);
			    	foreach(int f; 0..frames) // Downmix
			    		stereoHelper[0][f] = (stereoHelper[0][f]+stereoHelper[1][f]) * 0.5f;
					break;
				default:
				    dspModules[Module.Output].compute(frames, cast(float*[])dspBufferAccum, cast(float*[])outputs);
			}
		}

        /// Get access to the GUI
        if (OneTrickSimianGUI gui = cast(OneTrickSimianGUI) graphicsAcquire())
        {
            /// This is where you would update any elements in the gui
            /// such as feeding values to meters.
            
            gui.module1Indicator.setValueFromDSP(dspModules[Module.Kick].triggered()?1:0);
            gui.module2Indicator.setValueFromDSP(dspModules[Module.Snare].triggered()?1:0);
            gui.module3Indicator.setValueFromDSP(dspModules[Module.Rimshot].triggered()?1:0);
            gui.module4Indicator.setValueFromDSP(dspModules[Module.Clap].triggered()?1:0);
            gui.module5Indicator.setValueFromDSP(dspModules[Module.Tom1].triggered()?1:0);
            gui.module6Indicator.setValueFromDSP(dspModules[Module.Tom2].triggered()?1:0);
            gui.module7Indicator.setValueFromDSP(dspModules[Module.Tom3].triggered()?1:0);
            gui.module8Indicator.setValueFromDSP(dspModules[Module.HihatClosed].triggered()?1:0);
            gui.module9Indicator.setValueFromDSP(dspModules[Module.HihatOpen].triggered()?1:0);
            gui.module10Indicator.setValueFromDSP(dspModules[Module.Cymbal].triggered()?1:0);

            graphicsRelease();
        }
    }

    override IGraphics createGraphics()
    {
        return mallocNew!OneTrickSimianGUI(this);
    }
    
    ParamDescription getParamDesc(int index)
    {
	    return paramDesc[index];
    }

private:
	PresetBank _internalPresetBank;
	//float modWheel = 0;
	
    ParamDescription[paramMAX] paramDesc;
    float*[2] stereoHelper = [null, null];
    
    // Buffers for simian Modules
    float*[3] dspBufferTemp = [null, null, null];
    float*[3] dspBufferAccum = [null, null, null];

    PunkDSP[Module.Count] dspModules;
    
    void addBuffers(int frames, float*[] buffer, float*[] add)
    {
    	//dest[0..dest.length][0..frames] += src[0..dest.length][0..frames];
    	foreach (int chan; 0..3)
    	{
    		//for(int i=0; i<frames; i++)
    		foreach (int i; 0..frames)
    		{
    			buffer[chan][i] += add[chan][i];
    		}
    	}
    }
}

