/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

module punklabs.faust.common;

/* // Not needed...
import std.math;
import std.stdio;
import std.conv;
import std.algorithm;
import std.format;
import core.stdc.stdlib;
*/

import std.math;
import std.algorithm;

import dplug.core.nogc: mallocNew, mallocSlice, destroyFree, assumeNothrowNoGC;

import ldc.attributes; //fastmath

import dplug.core.vec;

alias FAUSTFLOAT = float;

class FastMath {
pure:
nothrow:
@nogc:
@trusted:
	//static @fastmath T pow(T) (T x, T n) { return x^^n; }
	/*
		Based on FastPow technique by Martin Ankerl
		https://martin.ankerl.com/2007/10/04/optimized-pow-approximation-for-java-and-c-c/
	*/
	static double pow(double base, double e) {
		union U {
			double d;
			int[2] i;
		}
		U u = { base };
		u.i[1] = cast(int)(e * (u.i[1] - 1072632447) + 1072632447);
		u.i[0] = 0;
		return u.d;
	}
	static double pow(double base, float e)
	{
		return pow(base, cast(double)e);
	} 
	static double pow(float base, double e)
	{
		return pow(cast(double)base, e);
	} 
	static float pow(float base, float e) {
		union U {
			float f;
			int i;
		}
		U u = { base };
		u.i = cast(int)(e * (u.i - 1065307417) + 1065307417);
		return u.f;
	}
	
	static float sqrt(float value) {
		union U {
			float f;
			int i;
		}
		U u = { value };
		u.i -= 1<<23;
		u.i = u.i >> 1;
		u.i += 1<<29;
		return u.f;
	}
	
}

class Meta {
nothrow:
@nogc:
    void declare(string name, string value) {}
}

class UI {
nothrow:
@nogc:
    void declare(string id, string key, string value) {}
    void declare(int id, string key, string value) {}
    void declare(FAUSTFLOAT* id, string key, string value) {}

    // -- layout groups

    void openTabBox(string label) {}
    void openHorizontalBox(string label) {}
    void openVerticalBox(string label) {}
    void closeBox() {}

    // -- active widgets

    void addButton(string label, FAUSTFLOAT* val) {}
    void addCheckButton(string label, FAUSTFLOAT* val) {}
    void addVerticalSlider(string label, FAUSTFLOAT* val, FAUSTFLOAT init, FAUSTFLOAT min, FAUSTFLOAT max, FAUSTFLOAT step) {}
    void addHorizontalSlider(string label, FAUSTFLOAT* val, FAUSTFLOAT init, FAUSTFLOAT min, FAUSTFLOAT max, FAUSTFLOAT step) {}
    void addNumEntry(string label, FAUSTFLOAT* val, FAUSTFLOAT init, FAUSTFLOAT min, FAUSTFLOAT max, FAUSTFLOAT step) {}

    // -- passive display widgets

    void addHorizontalBargraph(string label, FAUSTFLOAT* val, FAUSTFLOAT min, FAUSTFLOAT max) {}
    void addVerticalBargraph(string label, FAUSTFLOAT* val, FAUSTFLOAT min, FAUSTFLOAT max) {}
}

interface dsp {
nothrow:
@nogc:
public:
    void metadata(Meta* m);
    int getNumInputs();
    int getNumOutputs();
    void buildUserInterface(UI* uiInterface);
    int getSampleRate();
    void instanceInit(int sample_rate);
    void instanceResetUserInterface();
    void compute(int count, FAUSTFLOAT*[] inputs, FAUSTFLOAT*[] outputs);
    void initialize(int sample_rate);
}

/**
 * Implements and overrides the methods that would provide parameters for use in 
 * a plug-in or GUI.  These parameters are stored in a vector which can be accesed via
 * `readParams()`
 */
class FaustParamAccess : UI {
nothrow:
@nogc:
    this()
    {
        _faustParams = makeVec!FaustParam();
        //_faustParams.pushBack(FaustParam("ERROR", &missingValue, 0, 0, 0, 0, true, true));
        _faustParams.pushBack(FaustParam("ERROR", null, 0, 0, 0, 0, true, true));
    }
    ~this()
    {
    	_faustParams.destroy();
    }

    override void addButton(string label, FAUSTFLOAT* val)
    {
        _faustParams.pushBack(FaustParam(label, val, 0, 0, 0, 0, true));
    }
    
    override void addCheckButton(string label, FAUSTFLOAT* val)
    {
        _faustParams.pushBack(FaustParam(label, val, 0, 0, 0, 0, true));
    }
    
    override void addVerticalSlider(string label, FAUSTFLOAT* val, FAUSTFLOAT init, FAUSTFLOAT min, FAUSTFLOAT max, FAUSTFLOAT step)
    {
        _faustParams.pushBack(FaustParam(label, val, init, min, max, step));
    }

    override void addHorizontalSlider(string label, FAUSTFLOAT* val, FAUSTFLOAT init, FAUSTFLOAT min, FAUSTFLOAT max, FAUSTFLOAT step)
    {
        _faustParams.pushBack(FaustParam(label, val, init, min, max, step));
    }

    override void addNumEntry(string label, FAUSTFLOAT* val, FAUSTFLOAT init, FAUSTFLOAT min, FAUSTFLOAT max, FAUSTFLOAT step)
    {
        _faustParams.pushBack(FaustParam(label, val, init, min, max, step));
    }

    FaustParam[] readParams()
    {
    	return _faustParams.opSlice();
    }

    FaustParam readParam(int index)
    {
        return _faustParams[index];
    }
    
    FaustParam findParam(string label)
    {
    	int index = 0;
	    foreach(param; _faustParams)
	    {
	    	if(param.label == label)
	    		return _faustParams[index];
		    index++;
	    }
	    return _faustParams[0]; //Missing
    }


    ulong length()
    {
        return _faustParams.length();
    }

private:
	Vec!FaustParam _faustParams;
	//FAUSTFLOAT missingValue = 0;
}

struct FaustParam
{
	string label;
	FAUSTFLOAT* val;
	FAUSTFLOAT initial;
	FAUSTFLOAT min;
	FAUSTFLOAT max;
	FAUSTFLOAT step;
    bool isButton = false;
    bool isMissing = false;
}

class PunkDSP {
nothrow:
@nogc:
public:
	this(dsp _dsp)
	{
		this._dsp = _dsp;
		this.paramAccess = mallocNew!FaustParamAccess();
	}
	~this()
	{
		paramAccess.destroyFree();
		_dsp.destroyFree();
	}
	PunkDSP initializeParams()
	{
		if (!paramAccessInitialized)
		{
	        _dsp.buildUserInterface(cast(UI*)(&paramAccess));
	        paramAccessInitialized = true;
	        useWake("WakeUp");
	        useTrigger("Trigger");
	        useChoke("Choke");
		}
		return this;
	}
	PunkDSP trackSilence(float thresholdMS)
	{
		silenceThresholdMS = thresholdMS;
		initializeSilenceTracking();
		return this;
	}
	PunkDSP initialize(int sample_rate)
	{
		if (sample_rate != currentSampleRate)
		{
			currentSampleRate = sample_rate;
			_dsp.initialize(sample_rate);
			initializeParams();
			initializeSilenceTracking();
		}
		return this;
	}
	PunkDSP useWake(string label)
	{
		initializeParams();
        paramWake = paramAccess.findParam(label);
        if(paramWake.val)
			*(paramWake.val) = 0;
		return this;
	}
	PunkDSP useTrigger(string label)
	{
		initializeParams();
		paramTrigger = findParam(label);
		trigger(0);
		return this;
	}
	PunkDSP useChoke(string label)
	{
		initializeParams();
		paramChoke = findParam(label);
		choke(false);
		return this;
	}
    void metadata(Meta* m) { _dsp.metadata(m); }
    int getNumInputs() { return _dsp.getNumInputs(); }
    int getNumOutputs() { return _dsp.getNumOutputs(); }
    int getSampleRate() { return _dsp.getSampleRate(); }
    void compute(int frames, FAUSTFLOAT*[] inputs, FAUSTFLOAT*[] outputs)
    {
    	triggeredVelocityOnCompute = paramTrigger.val ? *(paramTrigger.val) : 0; //
    	_dsp.compute(frames, inputs, outputs);
    	trigger(0);
    	if(silenceThresholdSamples >= 0)
    	{
	        if(paramWake.val)
				*(paramWake.val) = 0;
    		int chans = cast(int)outputs.length;
    		if (chans==1)
    		{
				for(int i=0; i<frames; i+=silenceFrameSkip)
				{
					if (abs(outputs[0][i])>0.001) //Cycle Channels to coarsely estimate silence
						silenceCounter = 0;
					else
						silenceCounter++;
				}
    		}
    		else
    		{
				for(int i=0; i<frames; i+=silenceFrameSkip)
				{
					if (abs(outputs[i%chans][i])>0.001) //Cycle Channels to coarsely estimate silence
						silenceCounter = 0;
					else
						silenceCounter++;
				}
			}
    	}
    }
    void trigger(float velocity = 1)
    {
    	if(paramTrigger.val)
    		*(paramTrigger.val) = velocity;
    	if(velocity > 0)
	    	choke(false);
    }
    float triggeredVelocity()
    {
    	return triggeredVelocityOnCompute;
    }
    bool triggered()
    {
    	return (triggeredVelocity() > 0);
    }
    void choke(bool choked = true)
    {
    	if(paramChoke.val)
    	{
    		*(paramChoke.val) = choked ? 1 : 0;
    		if(choked)
    			trigger(0);
    	}
    }
    bool isSilent()
    {
    	return (silenceThresholdSamples >= 0 && silenceCounter >= silenceThresholdSamples);
    }
    void wakeUp()
    {
   		if (isSilent() && paramWake.val)
			*(paramWake.val) = 1;
		silenceCounter = 0;
    }
    FaustParam findParam(string label)
    {
    	return paramAccess.findParam(label);
	}

private:
	int currentSampleRate = 1;
	const int silenceFrameSkip = 4;
	dsp _dsp = null;
    FaustParamAccess paramAccess = null;
    bool paramAccessInitialized = false;
    
    float silenceThresholdMS = -1;
	int silenceCounter = 0;
	int silenceThresholdSamples = -1;//cast(int)(44100.0f*silenceThresholdMS/1000.0f);
	
	FaustParam paramTrigger;
	FaustParam paramChoke;
	FaustParam paramWake;
	
	float triggeredVelocityOnCompute = 0;

	
	void initializeSilenceTracking()
	{
		silenceThresholdSamples = cast(int)(cast(float)_dsp.getSampleRate() * silenceThresholdMS / 1000.0f) / silenceFrameSkip;
		silenceCounter = silenceThresholdSamples + 1; // Ensure we skip any initial smoothing
	}
}

