/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

module punklabs.ui.label;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;
import dplug.graphics.font;

import core.stdc.string;

import punklabs.ui.helper;
import punklabs.string.helper;

/// A simple label
class UILabel : UIElement
{
public:
nothrow:
@nogc:

	@ScriptProperty RGBA color = RGBA(128, 128, 128, 255);
	@ScriptProperty RGBA colorShadow = RGBA(0, 0, 0, 0);
	@ScriptProperty RGBA colorBackground = RGBA(0, 0, 0, 0);
	@ScriptProperty float shadowOffset = 0.1f;

	this(UIContext context, Font font, const(char)[] text, RGBA color = RGBA(255,255,255,255))
	{
		super(context, flagRaw);
		_font = font;
		setText(text);
		this.color = color;
	}
	~this()
	{
		if (_text !is null)
			StringHelper.freeString(_text.ptr);
	}

	override void onDrawRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects)
	{
		if (colorBackground.a > 0)
		{
			UIHelper.fillColorRaw(rawMap, dirtyRects, colorBackground);
		}
		if (_text !is null)
		{
			if (colorShadow.a > 0)
			{
				float y = position.size.y*shadowOffset;
				if (y != 0)
					UIHelper.drawLabelRaw(rawMap, dirtyRects, box2f.rectangle(0, y, position.size.x, position.size.y), StringHelper.getStringSlice(_text), _font, colorShadow);
			}
			UIHelper.drawLabelRaw(rawMap, dirtyRects, box2f.rectangle(0, 0, position.size.x, position.size.y), StringHelper.getStringSlice(_text), _font, color);
		}
	}

	void setColor(RGBA color)
	{
		if (this.color != color)
		{
			this.color = color;
			setDirtyWhole();
		}
	}
	void setBackgroundColor(RGBA color)
	{
		if(this.colorBackground != color)
		{
			this.colorBackground = color;
			setDirtyWhole();
		}
	}

	void setText(char* newText)
	{
		if (newText !is null)
			setText(newText[0..strlen(newText)]);
		else
			setText("");
	}
	void setText(const(char)[] newText)
	{
		/* // Allocate as normal
		if (_text !is null)
			StringHelper.freeString(_text.ptr);
		_text = StringHelper.duplicateString(newText);	
		//*/

		/* // Stress Test
		foreach(i; 0..1000)
		{
			if (_text !is null)
				StringHelper.freeString(_text.ptr);
			_text = StringHelper.duplicateString(newText);	
		}
		//*/
		
		//* // Avoid unnecessary allocations:
		if(_text is null)
		{
			_text = StringHelper.duplicateString(newText);
		}
		else
		{
			if(newText.length > strlen(_text.ptr))
			{
				StringHelper.freeString(_text.ptr);
				_text = StringHelper.duplicateString(newText);
			}
			else
			{
				strcpy(_text.ptr, newText.ptr);
			}
		}		
		//*/
		
		setDirtyWhole();
	}

private:
	Font _font;
    char[] _text = null;
	
}
