/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

module punklabs.ui.parameter;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;
//import dplug.wren;

import punklabs.ui.customgui;

/// A simple class to expose a parameter to Wren scripts
final class UIParameter : UIElement, IParameterListener
{
public:
nothrow:
@nogc:

	@ScriptProperty float floatValue;
	@ScriptProperty int intValue;
	@ScriptProperty float normalizedValue;

	this(UIContext context, Parameter param, UIElement invalidateElement = null)
	{
		super(context, invalidateElement ? (flagRaw|flagAnimated) : flagRaw);
		this.param = param;
		this.param.addListener(this);
		this.invalidateElement = invalidateElement;
		updateProperties();
	}
	~this()
	{
		param.removeListener(this);	
	}
	
	void updateProperties()
	{
		auto floatParam = cast(FloatParameter)param;
		if(floatParam)
		{
			normalizedValue = floatParam.getNormalized();
			floatValue = floatParam.value();
			intValue = cast(int)(floatValue + 0.5);
		}
		else
		{
			auto intParam = cast(IntegerParameter)param;
			if(intParam)
			{
				normalizedValue = intParam.getNormalized();
				intValue = intParam.value();
				floatValue = cast(float)intValue;
			}
		}
		updateNeeded = true;
		scriptMethodWasCalled = false;
	}

	override void onBeginParameterEdit(Parameter sender)
	{
	}
	override void onParameterChanged(Parameter sender)
	{
		atomicStore(__parameterChanged, true);
	}
	override void onEndParameterEdit(Parameter sender)
	{
	}
	
	override void onAnimate(double dt, double time)
	{
	    if (atomicLoad(__parameterChanged))
        {
			atomicStore(__parameterChanged, false);
			updateProperties();
		}
		if(updateNeeded)
		{
			updateNeeded = false;
			if (!scriptMethodWasCalled)
			{
				//if (scriptHasMethod && !context.wrenSupport.callPluginMethod("parameterChanged"))
				//	scriptHasMethod = false;
				auto rootElement = context.getRootElement();
				if (auto customRootElement = cast(ICustomGUI)rootElement) {
					customRootElement.parameterChanged();
				}
				else
				{
					rootElement.reflow();
				}
				
				scriptMethodWasCalled = true;
			}
		
			if(invalidateElement !is null)
				invalidateElement.setDirtyWhole();
		}
	}

private:
    shared(bool) __parameterChanged = false;  // access to this is through atomic ops
	Parameter param;
	UIElement invalidateElement = null;
	bool updateNeeded = false;
	shared static bool scriptHasMethod = true; //Until we learn otherwise
	shared static bool scriptMethodWasCalled = false;
}
